@@ -1,6 +1,7 @@
 #!/usr/bin/env python
 
 import datetime
+import hashlib
 import math
 import operator
 import optparse
@@ -13,20 +14,23 @@ import webbrowser
 from collections import namedtuple, OrderedDict
 from functools import wraps
 from getpass import getpass
+from io import TextIOWrapper
 
-# Py3k compat.
-if sys.version_info[0] == 3:
-    binary_types = (bytes, bytearray)
-    decode_handler = 'backslashreplace'
-    numeric = (int, float)
-    unicode_type = str
-    from io import StringIO
-else:
+# Py2k compat.
+if sys.version_info[0] == 2:
+    PY2 = True
     binary_types = (buffer, bytes, bytearray)
     decode_handler = 'replace'
     numeric = (int, long, float)
     unicode_type = unicode
     from StringIO import StringIO
+else:
+    PY2 = False
+    binary_types = (bytes, bytearray)
+    decode_handler = 'backslashreplace'
+    numeric = (int, float)
+    unicode_type = str
+    from io import StringIO
 
 try:
     from flask import (
@@ -64,6 +68,7 @@ else:
 
 from peewee import *
 from peewee import IndexMetadata
+from peewee import sqlite3
 from playhouse.dataset import DataSet
 from playhouse.migrate import migrate
 
@@ -97,7 +102,15 @@ ViewMetadata = namedtuple('ViewMetadata', ('name', 'sql'))
 class SqliteDataSet(DataSet):
     @property
     def filename(self):
-        return os.path.realpath(dataset._database.database)
+        db_file = dataset._database.database
+        if db_file.startswith('file:'):
+            db_file = db_file[5:]
+        return os.path.realpath(db_file.rsplit('?', 1)[0])
+
+    @property
+    def is_readonly(self):
+        db_file = dataset._database.database
+        return db_file.endswith('?mode=ro')
 
     @property
     def base_name(self):
@@ -177,14 +190,14 @@ class SqliteDataSet(DataSet):
 
 @app.route('/')
 def index():
-    return render_template('index.html')
+    return render_template('index.html', sqlite=sqlite3)
 
 @app.route('/login/', methods=['GET', 'POST'])
 def login():
     if request.method == 'POST':
         if request.form.get('password') == app.config['PASSWORD']:
             session['authorized'] = True
-            return redirect(session.get('next_url') or '/')
+            return redirect(session.get('next_url') or url_for('index'))
         flash('The password you entered is incorrect.', 'danger')
     return render_template('login.html')
 
@@ -546,12 +559,30 @@ def table_import(table):
                 format = 'json'
             else:
                 format = 'csv'
+
+            # Here we need to translate the file stream. Werkzeug uses a
+            # spooled temporary file opened in wb+ mode, which is not
+            # compatible with Python's CSV module. We'd need to reach pretty
+            # far into Flask's internals to modify this behavior, so instead
+            # we'll just translate the stream into utf8-decoded unicode.
+            if not PY2:
+                try:
+                    stream = TextIOWrapper(file_obj, encoding='utf8')
+                except AttributeError:
+                    # The SpooledTemporaryFile used by werkzeug does not
+                    # implement an API that the TextIOWrapper expects, so we'll
+                    # just consume the whole damn thing and decode it.
+                    # Fixed in werkzeug 0.15.
+                    stream = StringIO(file_obj.read().decode('utf8'))
+            else:
+                stream = file_obj.stream
+
             try:
                 with dataset.transaction():
                     count = dataset.thaw(
                         table,
                         format=format,
-                        file_obj=file_obj.stream,
+                        file_obj=stream,
                         strict=strict)
             except Exception as exc:
                 flash('Error importing file: %s' % exc, 'danger')
@@ -574,6 +605,7 @@ def drop_table(table):
     if request.method == 'POST':
         model_class = dataset[table].model_class
         model_class.drop_table()
+        dataset.update_cache()  # Update all tables.
         flash('Table "%s" dropped successfully.' % table, 'success')
         return redirect(url_for('index'))
 
@@ -665,6 +697,22 @@ def _close_db(exc):
     if not dataset._database.is_closed():
         dataset.close()
 
+
+class PrefixMiddleware(object):
+    def __init__(self, app, prefix):
+        self.app = app
+        self.prefix = '/%s' % prefix.strip('/')
+        self.prefix_len = len(self.prefix)
+
+    def __call__(self, environ, start_response):
+        if environ['PATH_INFO'].startswith(self.prefix):
+            environ['PATH_INFO'] = environ['PATH_INFO'][self.prefix_len:]
+            environ['SCRIPT_NAME'] = self.prefix
+            return self.app(environ, start_response)
+        else:
+            start_response('404', [('Content-Type', 'text/plain')])
+            return ['URL does not match application prefix.'.encode()]
+
 #
 # Script options.
 #
@@ -700,6 +748,17 @@ def get_option_parser():
         action='store_true',
         dest='prompt_password',
         help='Prompt for password to access database browser.')
+    parser.add_option(
+        '-r',
+        '--read-only',
+        action='store_true',
+        dest='read_only',
+        help='Open database in read-only mode.')
+    parser.add_option(
+        '-u',
+        '--url-prefix',
+        dest='url_prefix',
+        help='URL prefix for application.')
     return parser
 
 def die(msg, exit_code=1):
@@ -726,9 +785,38 @@ def install_auth_handler(password):
         if not session.get('authorized') and request.path != '/login/' and \
            not request.path.startswith(('/static/', '/favicon')):
             flash('You must log-in to view the database browser.', 'danger')
-            session['next_url'] = request.path
+            session['next_url'] = request.base_url
             return redirect(url_for('login'))
 
+def initialize_app(filename, read_only=False, password=None, url_prefix=None):
+    global dataset
+    global migrator
+
+    if password:
+        install_auth_handler(password)
+
+    if read_only:
+        if sys.version_info < (3, 4, 0):
+            die('Python 3.4.0 or newer is required for read-only access.')
+        if peewee_version < (3, 5, 1):
+            die('Peewee 3.5.1 or newer is required for read-only access.')
+        db = SqliteDatabase('file:%s?mode=ro' % filename, uri=True)
+        try:
+            db.connect()
+        except OperationalError:
+            die('Unable to open database file in read-only mode. Ensure that '
+                'the database exists in order to use read-only mode.')
+        db.close()
+        dataset = SqliteDataSet(db, bare_fields=True)
+    else:
+        dataset = SqliteDataSet('sqlite:///%s' % filename, bare_fields=True)
+
+    if url_prefix:
+        app.wsgi_app = PrefixMiddleware(app.wsgi_app, prefix=url_prefix)
+
+    migrator = dataset._migrator
+    dataset.close()
+
 def main():
     # This function exists to act as a console script entry-point.
     parser = get_option_parser()
@@ -738,25 +826,27 @@ def main():
 
     password = None
     if options.prompt_password:
-        while True:
-            password = getpass('Enter password: ')
-            password_confirm = getpass('Confirm password: ')
-            if password != password_confirm:
-                print('Passwords did not match!')
-            else:
-                break
+        if os.environ.get('SQLITE_WEB_PASSWORD'):
+            password = os.environ['SQLITE_WEB_PASSWORD']
+        else:
+            while True:
+                password = getpass('Enter password: ')
+                password_confirm = getpass('Confirm password: ')
+                if password != password_confirm:
+                    print('Passwords did not match!')
+                else:
+                    break
 
-    if password:
-        install_auth_handler(password)
+    # Initialize the dataset instance and (optionally) authentication handler.
+    initialize_app(args[0], options.read_only, password, options.url_prefix)
 
-    db_file = args[0]
-    global dataset
-    global migrator
-    dataset = SqliteDataSet('sqlite:///%s' % db_file, bare_fields=True)
-    migrator = dataset._migrator
-    dataset.close()
     if options.browser:
         open_browser_tab(options.host, options.port)
+
+    if password:
+        key = b'sqlite-web-' + args[0].encode('utf8') + password.encode('utf8')
+        app.secret_key = hashlib.sha256(key).hexdigest()
+
     app.run(host=options.host, port=options.port, debug=options.debug)
 
 
